# Date/time utilities
# Jim Davis 1 Aug 1994

# Expects a line of
# the form [DAY] MONTH DATE YEAR
# Return {Day Month Date Year} where
#   Day is 0 to 6 or "" if none supplied
#   Month is a number 0 to 11.
# There is no obvious way to pass strings with embedded whitespace on the command line.
# So we allow you to use a hyphen to separate components.

proc parse_date {string} {
    set day ""
    if [regexp -nocase "(\[0-9\]\[0-9\]?)/(\[0-9\]\[0-9\]?)/(\[0-9\]\[0-9\])" $string mv month date year] {
    } else {
	if [regexp -nocase "^(\[A-Za-z\]+) *(.*)\$" $string mv day_name rest] {
	    set day [day_number $day_name]
	    if {$day >= 0} {
		set string $rest
	    } else {set day ""}}
	if [regexp -nocase "(\[A-Za-z\]+)\[-, \]*(\[0-9\]+)\[-, \]+(\[0-9\]+)" $string mv month date year] {
	    set month [month_number $month]
	} else {
	    if [regexp -nocase "(\[0-9\]+)\[-, \]+(\[A-Za-z\]+)\[-, \]*(\[0-9\]+)" $string mv date month year] {
	    set month [month_number $month]
	} else {
	    error "Could not parse $string as a date"}}}

    if {$year < 100} {
	if {$year < 50} {set year [expr {2000 + $year}]} {set year [expr {1900 + $year}]}}

    set date [string trimleft $date 0]
    list $day $month $date $year}
    

set short_day_names {Sun Mon Tue Wed Thu Fri Sat}
set long_day_names {Sunday Monday Tuesday Wednesday Thursday Friday Saturday}

proc day_number {day} {
    global short_day_names long_day_names
    set temp [lsearch $short_day_names $day]
    if {$temp >= 0} {return $temp}
    set temp [lsearch $long_day_names $day]
    if {$temp >= 0} {return $temp}
    return -1}


set short_month_names {Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec}
set long_month_names {January February March April May June July August September October November December}

proc month_number {month} {
    global short_month_names long_month_names
    set temp [lsearch $short_month_names $month]
    if {$temp >= 0} {return $temp}
    set temp [lsearch $long_month_names $month]
    if {$temp >= 0} {return $temp}
    error "Can't understand month name $month"}


# compare two parsed dates
# return -1, 0 or 1 as date1 is before, same, or after date2
proc date_compare {date1 date2} {
    set mon1 [lindex $date1 1]
    set day1 [lindex $date1 2]
    set year1 [lindex $date1 3]

    set mon2 [lindex $date2 1]
    set day2 [lindex $date2 2]
    set year2 [lindex $date2 3]

    if {$year1 < $year2} {return -1}
    if {$year1 > $year2} {return 1}
    # years are equal

    if {$mon1 < $mon2} {return -1}
    if {$mon1 > $mon2} {return 1}
    # months are equal

    if {$day1 < $day2} {return -1}
    if {$day1 > $day2} {return 1}
    # days are equal

    return 0}

set days_per_month  {31 28 31 30 31 30 31 31 30 31 30 31}

proc month_ndays {month year} {
    global days_per_month
    if {[leap_year_p $year] && $month == 1} {
	return 29
    } else {lindex $days_per_month $month}}

proc leap_year_p {year} {
    expr {$year % 4 == 0 && $year % 100 != 0}}

proc date_plus {parsed_date days} {
    set day [lindex $parsed_date 0]
    set month [lindex $parsed_date 1]
    set date [lindex $parsed_date 2]
    set year [lindex $parsed_date 3]

    if {$day != ""} {
	set day [expr {($day + $days) % 7}]}

    if {$days > 0} {
	incr date $days
	while {$date > [month_ndays $month $year]} {
	    incr date -[month_ndays $month $year]
	    incr month
	    if {$month == 12} {
		set month 0
		incr year}}}
    if {$days < 0} {
	incr date $days
	while {$date <= 0} {
	    incr month -1
	    if {$month == -1} {
		set month 11
		incr year -1}
	    incr date [month_ndays $month $year]}}

    list $day $month $date $year}

proc date_string {parsed_date} {
    global short_month_names short_day_names
    set day  [lindex $parsed_date 0]
    format "%s%s-%d-%d" \
	[expr {($day == {}) ? "" : [join [list [lindex $short_day_names $day] "-"] ""]}] \
	[lindex $short_month_names [lindex $parsed_date 1]] \
	[lindex $parsed_date 2] \
	[lindex $parsed_date 3]}


