#!/usr/local/bin/tclsh

# Analyze results in .journal file
# Jim Davis Aug 23 1993
# 1 Aug 94 modified to take start and end date arguments, and to print
#  reports by day... in other words, its good for timecards.
# 8 Aug 94 to use tclsh instead of wish so no display is required.
# 21 Aug 94 minor formatting glitches

# variables used:
# filename - the file to read from
# consumed - array.  index is category name, holds number of minutes
# ndays - number of days for which data exists
# first_day - first date for which data exists
# last_day - last date for which data exists
# ignores - list of category names not be included in report.


# Default values for command line args
set ignores {}
set filename $env(HOME)/.timeclock.times
set start_date ""
set end_date ""
set by_days 0
set print_cumulative 0
set zeroes 0
set round 0
# array mapping cat to new name
set renames("") ""
set current_day ""

set helpinfo "

Syntax:  time-report.tcl \[options\] \[FILENAME\]

Use this command to analyze the data accumulated by the timeclock
command, which creates a file recording how you have spent your time
in the office.

Options:

-ignore CATEGORY
  Excludes category from statistics e.g. if you want to exclude time
  you were sick or playing Go.  You can supply this argument as many
  times as you want.

-start DATE
 first date to use, in form Mon-DD-YYYY

-end DATE
 last date to use, in form Mon-DD-YYYY

-by_days BOOLEAN
 list time used by days as well as summing.  Default 0

-rename CATEGORY NAME
  Can be used more than once.  Useful for summing categories

-zeroes BOOLEAN
  Include categories that are zero.  Default 0

-round BOOLEAN
  round to nearest 1/2 hour.  

FILENAME is the name of the file holding the data.  By default it's $filename
"


proc parse_command_line {} {
    global argv
    set nargs [llength $argv]
    for {set i 0} {$i < $nargs} {incr i} {
	set arg [lindex $argv $i]
	case $arg in {
	    {help} {global helpinfo;puts $helpinfo ; exit}
	    {-ignore} {
		global ignores
		lappend ignores [get_command_arg $arg]}
	    {-start} {
		global start_date
		set start_date [parse_date [get_command_arg $arg]]}
	    {-end} {
		global end_date
		set end_date [parse_date [get_command_arg $arg]]}
	    {-by_days} {
		global by_days
		set by_days [get_command_arg $arg]}
	    {-zeroes} {
		global zeroes
		set zeroes [get_command_arg $arg]}
	    {-round} {
		global round
		set round [get_command_arg $arg]}
	    {-rename} {
		global renames
		set cat [get_command_arg $arg]
		set name [get_command_arg $arg]
		set renames($cat) $name}
	    -* {
		puts stderr "unrecognized arg $arg"; exit}
	    default {
		global filename
		set filename $arg}
	}}}

proc get_command_arg {name} {
    upvar 1 i i
    upvar 1 nargs nargs
    upvar 1 argv argv
    incr i
    if {$i == $nargs} {puts stderr [format "Missing value for %s" $name]; exit}
    lindex $argv $i}



proc read_journal_file {file} {
    global ndays first_day last_day consumed ignores by_days
    catch {unset consumed}
    set handle [open $file]
    set line_count 0
    set ndays 0
    set state none
    set interested 0
    set first_day "no first date in file?!"
    set last_day "no last date in file?!"
    for {set cl [gets $handle line]} {$cl>=0} {set cl [gets $handle line]} {
        incr line_count
	case $state in {
	    none {
		if {$cl>0} {
		    # non-blank line must be a date, the start of a new record.
		    if [interesting_date_p $line] {
			if {$ndays==0} {set first_day $line} {set last_day $line}
			incr ndays
			set state day
			if $by_days {day_begin $line}
		    } else {set state skip}}}
	    day {
		if {$line==""} {
		    # blank line is end of record.
		    finish_day
		    set state none
		} else {
		    if [regexp "^(.*) (\[0-9\]+)\$" $line mv category minutes] {
			if [expr [lsearch $ignores $category]<0] {
			day_cat $category $minutes}
		    } else {
			error [format "bad syntax in \"%s\" (line %d)" $line $line_count]}
		    }}
	    skip {
		# skip lines from uninteresting days
		if {$line==""} {
		    # end of record
		    set state none}
	    }

	}}
    finish_day
    }

proc interesting_date_p {date} {
    global start_date end_date
    set date [parse_date $date]
    expr {($start_date == "" || [date_compare $date $start_date] > -1) && \
	($end_date == "" || [date_compare $date $end_date] < 1)}}

proc day_begin {day} {
    global current_day
    global day_cats; set day_cats ""
    set current_day $day}

proc day_cat {category minutes} {
    global renames
    if {[lsearch [array names renames] $category] >= 0} {
	set category $renames($category)}
    global consumed by_days
    if [catch {set ignore $consumed($category)}] {
	set consumed($category) 0}
    incr consumed($category) $minutes

    if $by_days {
	global day_cats
	lappend day_cats [list $category $minutes]
    }
}

proc finish_day {} {
    global current_day day_cats day_info
    if {$current_day != ""} {
	lappend day_info [list $current_day $day_cats]}
    set current_day ""
    set day_cats {}}


proc print_results {} {
    global by_days print_cumulative

    global ndays first_day last_day 
    puts [format "File covers %d working days, from %s to %s" $ndays $first_day $last_day]

    # Sum up total time consumed into total_time
    # Make a list of pairs [CATEGORY TIME] into cat_totals

    global consumed  zeroes
    set names [array names consumed]
    global cat_totals

    set total_time 0
    set cat_totals {}
    set id [array startsearch consumed]
    for {set category [array nextelement consumed $id]} \
	{$category!=""} \
	{set category [array nextelement consumed $id]} {
	    if {$consumed($category) > 0 || $zeroes} {
		lappend cat_totals [list $category  $consumed($category)]
		incr total_time $consumed($category)}}
    array donesearch consumed $id

    set hours [expr $total_time/60]
    set minutes [expr $total_time%60]
    puts [format "Total: %d hours %d minutes" $hours $minutes]

    if $by_days {print_by_days $cat_totals}
    if $print_cumulative {
	puts "" ; puts ""

	print_cumulative $cat_totals $total_time}
}


proc print_by_days {cat_totals} {
    global day_info zeroes round
    set cats [mapcar first $cat_totals]
    set width [max 5 [lmax [mapcar string-length $cats]]]

    # list of day names and list of dates
    set days {}
    set dates {}
    global short_day_names
    foreach di $day_info {
	set day [parse_date [first $di]]
	lappend days [lindex $short_day_names [lindex $day 0]]
	lappend dates [join [list [expr [lindex $day 1]+1] [lindex $day 2]] "/"]}

    # 5 is number of chars needed to print HH:MM
    set day_width [max 5 [lmax [mapcar string-length $dates]]]

    # print Day names
    puts -nonewline [format "%${width}s " ""]
    foreach day $days {
	puts -nonewline [format "%${day_width}s " $day]}
    puts ""

    # print Dates
    puts -nonewline [format "%${width}s " ""]
    foreach date $dates {
	puts -nonewline [format "%${day_width}s " $date]}
    puts "Total"

    # for each cat, print amount used each day

    global cats_by_day
    set cats_by_day {}
    set day_index 0
    foreach di $day_info {set day_total($day_index) 0; incr day_index}
    foreach cat $cats {
	set temp {}
	set weekly 0
	set day_index 0
	foreach di $day_info {
	    set amt 0
	    # A category might appear more than once in a day 
	    # if renaming occured
	    foreach sum [lsecond $di] {
		if {[first $sum] == $cat} {incr amt [lsecond $sum]}}
	    if $round {
		set amt [round $amt 30]}
	    incr weekly $amt
	    incr day_total($day_index) $amt
	    lappend temp $amt
	    incr day_index}
	lappend cats_by_day [list $cat $temp $weekly]}
	
    foreach ci $cats_by_day {
	set cat [first $ci]
	puts -nonewline [format "%-${width}s " $cat]
	foreach di [lsecond $ci] {
	    puts -nonewline [format "%${day_width}s " [min-to-hrs-and-min $di]]}
	puts [format "%${day_width}s " [min-to-hrs-and-min [lthird $ci]]]}

    puts -nonewline [format "%-${width}s " "Total"]
    set day_index 0
    set week_total 0
    foreach di $day_info {
	puts -nonewline [format "%${day_width}s " [min-to-hrs-and-min $day_total($day_index)]]
	incr week_total $day_total($day_index)
	incr day_index}
    puts [format "%${day_width}s " [min-to-hrs-and-min $week_total]]

}
    

proc round {n nearest} {
    expr {( $n / $nearest + ( (($n%$nearest) > ($nearest/2)) ? 1 : 0)) * $nearest}}

proc min-to-hrs-and-min {minutes} {
    format "%2d:%02d"  [expr $minutes/60] [expr $minutes%60]}

proc print_cumulative {list total} {
    set list [sort $list 1]
    puts [format "%-20s %7s %4s %s" category time % "cum. %"]
    set accumulator 0
    set pct_acc 0
    foreach elt $list {
	set category [lindex $elt 0]
	set minutes [lindex $elt 1]
	incr accumulator $minutes
	set pct [expr $minutes*100.0/$total]
	set pct_acc [expr $pct_acc+$pct]
	puts [format "%-20s %4d:%02d %4.1f %5.1f" \
	      $category [expr $minutes/60] [expr $minutes%60] \
		  $pct	   \
		  [expr $accumulator*100.0/$total] ]}
    puts [format  "%-20s %4d:%02d %5.1f" \
	  "Total" [expr $accumulator/60] [expr $accumulator%60] $pct_acc]

}

# list is a list of elements, each of them a list.  key is the index in those
# elements of a numeric field.  result is a new list containing the elements,
# sorted in decreasing numeric order by the key field.
# no apology for simplistic implementation
proc sort {list {key 0}} {
    set result {}
    for {} {[llength $list]>0} {} {
	set max 0
	set maxi -1
	set i 0
	foreach elt $list {
	    set v [lindex $elt $key]
	    if {$maxi < 0 || $v > $max} {
		set maxi $i
		set max $v}
	    incr i}
	lappend result [lindex $list $maxi]
	set list [lreplace $list $maxi $maxi]
    }
    return $result}

# Some utilities

# This is a copy of the max from utils.tcl
proc max {args} {
   set x [lindex $args 0]
   foreach i $args {
      if {$i>$x} {set x $i}
   }
   return $x
}

proc lsecond {l} {return [lindex $l 1]}
proc lthird {l} {return [lindex $l 2]}


parse_command_line 
#wm withdraw .
puts  "Analyzing data in $filename"
read_journal_file $filename
print_results
#update

#exit 0
