#!/usr/local/X11/bin/wish -f

# Program to keep track of time usage
# Jim Davis, Design Research Institute, Cornell Univ, Feb 93
# 18 Apr 94 new .timeclock file version 1 allows spaces in names

set myname {Time Clock}
wm title . $myname
wm iconname . $myname

option add *background honeydew

# font for clocks
#set rfont  {-*-clean-medium-r-normal-*-12-*-*-*-*-*-*-*}
# use fixed width font so we can align things.
set rfont  {-b&h-*-medium-r-normal-*-10-*-*-*-m-*-*-*}
# font for menu buttons
set mfont  {-*-clean-medium-r-normal-*-8-*-*-*-*-*-*-*}

set helptext "This program tracks the accumulated time you spend on various activities.  If it were smarter, it would look over your shoulder and watch you, but instead you have to tell it what you are doing by selecting one of the radio buttons in the window.\nThe list of clocks it keeps are defined by a file .timeclock in your home directory.  This file must begin with the line TIMECLOCK 0.  Remaining lines have the name of a clock and the accumulated time in minutes.  (For an example, look at mine.)  At the end of the day, click on Home.  This writes to file .timeclock.times the accumulated time for the day, then clears the clocks.\nOptions has additional options.\n\nSuggestions welcome.\n
Written by Jim Davis Feb 1993\ndavis@dri.cornell.edu"


## Default values 

# Where clock data is stored
set clock_file $env(HOME)/.timeclock

# catgories of actions
set default_categories {meeting demo lunch other}

# Parse command line args - none yet

set args [split $argv]
set nargs [llength $args]
for {set i 0} {$i < $nargs} {incr i} {
 set arg [lindex $args $i]
 case $arg in {
  default {puts stderr [format "unrecognized arg %s\n" $arg];exit}
  }}


## state variables

# categories is a list of names, defined below
# times is an array of accumulated time in minutes, defined below.

# The clock currently ticking - -1 or else an index into array times
set current_clock -1

## User interface

# Make radio buttons and clock displays for the categories
proc make-fields {} {
    global color rfont categories times
    catch {destroy .clocks}
    frame .clocks 
    set place {top expand fillx frame w}
    set max_width 0
    foreach name $categories {
	if {[string length $name] > $max_width} {set max_width [string length $name]}}
    incr max_width
    set fmtstring %-${max_width}s
    set index 0
    foreach name $categories {
	set line .clocks.[string tolower $name]
	frame $line 
	radiobutton $line.button  \
	    -relief flat -font $rfont \
	    -value $name \
	    -command [list select-clock $index] \
	    -text [format $fmtstring $name]
	message $line.clock  \
	    -font $rfont -aspect 1000\
	    -text [clock-string $times($index)]
	pack append $line $line.button {left} $line.clock {left expand fillx}
	if $index==0 { pack append .clocks $line $place} else {pack after $prev $line $place}
	set prev $line
	incr index}
}

# Buttons to control operation
frame .controls 
button .controls.quit  -font $mfont -text Quit -command {destroy .}
button .controls.options  -font $mfont -text Options -command {clock-options }
button .controls.go-home   -font $mfont -text {Home} -command {go-home}
button .controls.help  -font $mfont -text Help \
    -command {show-help-window $myname $helptext}

pack append .controls \
    .controls.quit {left} \
    .controls.options {left} \
    .controls.go-home {left} \
    .controls.help {left}

# Click here when you are not in the office.
radiobutton .absent  \
	    -relief flat -font $rfont \
	    -command {select-clock -1} \
	    -text Absent



proc select-clock {clock} {
    global current_clock
    set current_clock $clock
}

# String specifying elapsed time in days (8 hour days!), hours, minutes
proc clock-string {minutes} {
    set minus ""
    if {$minutes<0} {set minus "-"; set minutes [expr {- $minutes}]}
    set days [expr {$minutes / 480}]
    set minutes [expr {$minutes - $days * 480}]
    set hours [expr {$minutes / 60}]
    set minutes [expr {$minutes - $hours * 60}]
    if {$days>1} {
	set day_string [format "%d days " $days]
    } else {
	if {$days==1} {
	    set day_string "1 day "
	} else {
	    set day_string ""}}
    format "%s%s%d:%02d" $minus $day_string $hours $minutes
 }


# Tick rate is once per minute, which is 60000 msec
proc tick {} {
    after 60000 tick
    global current_clock times categories
    if $current_clock>=0 {
	incr times($current_clock)
	set name [string tolower [lindex $categories $current_clock]]
	.clocks.$name.clock configure -text [clock-string $times($current_clock)]
    }
}

proc clear-clocks {} {
    global categories times
    set i 0
    foreach name $categories {
	set times($i) 0
	.clocks.[string tolower $name].clock configure -text [clock-string $times($i)]
	incr i}
}

    
proc stop-clocks {} {
    .absent select
    select-clock -1}


# File I/O

set timeclock_version_string {TIMECLOCK 1}

# Save current running clock values
proc write-clocks-file {pathname} {
    global categories times timeclock_version_string
    set handle [open $pathname w]
    puts $handle $timeclock_version_string
    set i 0
    foreach name $categories {
	puts $handle [format "%s: %d" $name $times($i)]
	incr i}
    close $handle
}

proc load-clocks-file {pathname} {
    global categories times timeclock_version_string
    set handle [open $pathname r]
    gets $handle line
    set i 0
    if [string match $line {TIMECLOCK 0}] {
	set categories {}
	catch {unset times}
	for {set c [gets $handle line]} {$c>0} {set c [gets $handle line]} {
	    if {[string index $line 0]!= [string index "#" 0]} {
		set fields [split $line]
		set name [lindex $fields 0]
		set minutes [lindex $fields 1] 
		lappend categories $name
		set times($i) $minutes
		incr i
	    }}} else {if [string match $line {TIMECLOCK 1}] {
	set categories {}
	catch {unset times}
	for {set c [gets $handle line]} {$c>0} {set c [gets $handle line]} {
	    if {[string index $line 0]!= [string index "#" 0]} {
		set fields [split $line ":"]
		set name [lindex $fields 0]
		set minutes [lindex $fields 1] 
		lappend categories $name
		set times($i) $minutes
		incr i
	    }}} 	else {puts stderr "timeclock file does not begin TIMECLOCK 0"}}}

# use this if no .timeclock
proc make-default-categories {} {
    global categories times default_categories
    catch {unset times}
    set categories $default_categories
    set i 0
    foreach name $default_categories {
	set times($i) 0
	incr i}}

proc load-clock-file-or-default {} {
    global clock_file
    if [file exists $clock_file] {
	load-clocks-file $clock_file
    } else {
	puts stderr "No file $clock_file, using default values"
	make-default-categories}}


## Journal file - time used each day

set journal_file $env(HOME)/.timeclock.times

proc write-journal-file {pathname} {
    global categories times 
    set handle [open $pathname [expr {[file exists $pathname]?{a+}:{w}}]]
    set date [exec date]
    puts $handle [format "%s %s %s %s" [lindex $date 0] [lindex $date 1] [lindex $date 2] [lindex $date 5]]
    set i 0
    foreach name $categories {
	puts $handle [format "%s %d" $name $times($i)]
	incr i}
    puts $handle ""
    close $handle
}

proc go-home {} {
    stop-clocks
    global journal_file
    write-journal-file $journal_file
    clear-clocks
}


menu .options -background white -font $mfont
bind .options <Leave> {.options unpost}
.options add command -label {Adjust time allocation} -command {edit-time}
.options add command -label {Save (write .timeclock file)} -command {write-clocks-file $clock_file}

proc clock-options {} {
    .options post \
	[winfo rootx .controls.options] \
	[expr {[winfo rooty .controls.options] + [winfo height .controls.options]}]}



set edit_helptext "This window allows you to change the allocation of time.  The - button removes 5 minutes, the + adds 5.  Removed time is transfered to \"Extra\".  Click Done when done, or Cancel to cancel."

proc edit-time {} {
    global rfont mfont categories times saved_times extra

    toplevel .edit 
    wm title .edit {Edit time allocation}

    set max_width 0
    foreach name $categories {
	if {[string length $name] > $max_width} {set max_width [string length $name]}}
    incr max_width
    set fmtstring %-${max_width}s

    catch {unset saved_times}
    set index 0
    foreach name $categories {
	set saved_times($index) $times($index)
	incr index}
    set extra 0

    set place {top expand fillx frame w}
    set index 0
    foreach name $categories {
	set line .edit.[string tolower $name]
	frame $line 
	message $line.name  \
	      -relief flat -font $rfont -aspect 1000 \
	    -text [format $fmtstring $name]
	button $line.decf   \
	    -text "-" -command [list edit-decf $index]
	message $line.clock  \
	    -font $rfont -aspect 1000\
	    -text [clock-string $saved_times($index)]
	button $line.incf   \
	    -text "+" -command [list edit-incf $index]
	pack append $line \
	    $line.name {left} \
	    $line.decf  {left expand fillx}\
	    $line.clock {left expand fillx}\
	    $line.incf {left expand fillx}
	pack append .edit $line $place
	set prev $line
	incr index}

# "extra" line at bottom
    set line .edit.extra
    frame $line 
    message $line.name \
	  -relief flat -font $rfont -aspect 1000 \
	-text [format $fmtstring "Extra"]
    button $line.decf   \
	    -text "-" -state disabled
    message $line.clock  \
	-font $rfont -aspect 1000\
	-text [clock-string 0]
    button $line.incf   \
	-text "+" -state disabled
    pack append $line \
	$line.name {left} \
	$line.decf  {left expand fillx}\
	$line.clock {left expand fillx}\
	$line.incf {left expand fillx}

    pack append .edit \
	$line $place

# and finally a frame of control buttons

    frame .edit.controls 
    button .edit.controls.done   -font $mfont \
	-text Done -command {edit-done}
    button .edit.controls.cancel   -font $mfont \
	-text {Cancel} -command {edit-cancel}
    button .edit.controls.help  -font $mfont \
	-text Help  -command {show-help-window {time clock edit} $edit_helptext "" {}}
    pack append .edit.controls \
	.edit.controls.done {left} \
	.edit.controls.cancel {left} \
	.edit.controls.help {left}
    pack append .edit \
	.edit.controls {top expand fillx}
}

proc edit-decf {index} {
    global saved_times extra 
    set amt $saved_times($index)
    if $amt<5 {
	set saved_times($index) 0
	incr extra $amt
    } else {
	set saved_times($index) [expr {$amt - 5}]
	incr extra 5}
    edit-clock-changed $index
}

proc edit-incf {index} {
    global saved_times extra 
    set amt $saved_times($index)
    set saved_times($index) [expr {$amt + 5}]
    incr extra -5
    edit-clock-changed $index
}

proc edit-clock-changed {index} {
    global saved_times categories extra
    .edit.[string tolower [lindex $categories $index]].clock configure -text [clock-string $saved_times($index)]
    .edit.extra.clock configure -text [clock-string $extra]
}

proc edit-done {} {
    global categories times saved_times
    set index 0
    foreach name $categories {
	set times($index) $saved_times($index)
	.clocks.[string tolower $name].clock configure -text [clock-string $times($index)]
	incr index}
    destroy .edit
}

proc edit-cancel {} {
    destroy .edit
}

## top level setup

load-clock-file-or-default
make-fields 

# toplevel frame
pack append . \
    .absent {top expand fillx} \
    .clocks {top expand fillx} \
    .controls {top expand fillx}

tick

# todo
# code to periodically save file (in case of crash)
# UI code to add/delete category
# code to allow another application to tell us what user's doing
